// SPDX-License-Identifier: GPL-2.0
/*
 * Copyright (c) 2000-2005 Silicon Graphics, Inc.
 * All Rights Reserved.
 */

#ifndef __LIBXCMD_H__
#define __LIBXCMD_H__

#include "libxfs.h"
#include <sys/time.h>

#endif	/* __LIBXCMD_H__ */
