/* SPDX-License-Identifier: GPL-2.0-or-later */
/*
 * Copyright (c) 2019 Oracle, Inc.
 * All Rights Reserved.
 */
#ifndef __LIBFROG_LOGGING_H__
#define __LIBFROG_LOGGING_H__

void xfrog_perror(int error, const char *s);

#endif	/* __LIBFROG_LOGGING_H__ */
