// SPDX-License-Identifier: GPL-2.0
/*
 * Copyright (c) 2000-2001,2005 Silicon Graphics, Inc.
 * All Rights Reserved.
 */

#define	DEBUG_FLIST	0x1

extern long	debug_state;
extern void	debug_init(void);
