// SPDX-License-Identifier: GPL-2.0
/*
 * Copyright (c) 2000-2001,2005 Silicon Graphics, Inc.
 * All Rights Reserved.
 */

extern const struct field	agi_flds[];
extern const struct field	agi_hfld[];

extern void	agi_init(void);
extern int	agi_size(void *obj, int startoff, int idx);
