/* SPDX-License-Identifier: GPL-2.0 */
/*
 * Copyright (c) 2000-2005 Silicon Graphics, Inc.
 * All Rights Reserved.
 */
#ifndef __LIBFROG_CRC32C_H__
#define __LIBFROG_CRC32C_H__

extern uint32_t crc32c_le(uint32_t crc, unsigned char const *p, size_t len);

#endif /* __LIBFROG_CRC32C_H__ */
