/* SPDX-License-Identifier: GPL-2.0 */
/*
 * Copyright (c) 2024 Christoph Hellwig.
 */
#ifndef _XFS_REPAIR_ZONED_H_
#define _XFS_REPAIR_ZONED_H_

void check_zones(struct xfs_mount *mp);

#endif /* _XFS_REPAIR_ZONED_H_ */
