// SPDX-License-Identifier: GPL-2.0
/*
 * Copyright (c) 2000-2001,2005 Silicon Graphics, Inc.
 * All Rights Reserved.
 */

extern int		blkbb;
extern int		exitcode;
extern int		expert_mode;
extern xfs_mount_t	*mp;
extern struct libxfs_init x;
extern xfs_agnumber_t	cur_agno;
