// SPDX-License-Identifier: GPL-2.0+
/*
 * Copyright (C) 2017 Oracle.  All Rights Reserved.
 * Author: Darrick J. Wong <darrick.wong@oracle.com>
 */
extern void	fuzz_init(void);
extern void	fuzz_struct(const field_t *fields, int argc, char **argv);
