// SPDX-License-Identifier: GPL-2.0
/*
 * Copyright (c) 2000-2001,2005 Silicon Graphics, Inc.
 * All Rights Reserved.
 */

extern int	dbprefix;

extern int	dbprintf(const char *, ...);
extern void	logprintf(const char *, ...);
extern void	output_init(void);
