// SPDX-License-Identifier: GPL-2.0
/*
 * Copyright (c) 2012 Red Hat, Inc.
 * All Rights Reserved.
 */
#ifndef XFS_SPACEMAN_INIT_H_
#define XFS_SPACEMAN_INIT_H_

extern char	*progname;
extern int	exitcode;

#endif /* XFS_SPACEMAN_INIT_H_ */
