// SPDX-License-Identifier: GPL-2.0
/*
 * Copyright (c) 2000-2001,2005 Silicon Graphics, Inc.
 * All Rights Reserved.
 */

extern void	blockint(void);
extern void	clearint(void);
extern void	init_sig(void);
extern int	seenint(void);
extern void	unblockint(void);
